# Ecto

## Ecto Setup

- За да си инсталираме Ecto ни трябва `ecto_sql` и адаптер.
- За адаптер ще ползваме postgres, но има [много други](https://www.bigdatamark.com/list-of-available-ecto-sql-adapters-for-elixir).
- Това са ни началните `deps`:

```elixir
  defp deps do
    [
      {:ecto_sql, "~> 3.10"},
      {:postgrex, ">= 0.0.0"}
    ]
  end
```


### Ecto Setup

- Промените са да добавим `Books.Repo` като дете на Supervisor-a и да го конфигурираме:

```elixir
# Books.Application

children = [
  Books.Repo
]
```


### Ecto Setup

- Промените са да добавим `Books.Repo` като дете на Supervisor-a и да го конфигурираме:

```elixir
# config/config.exs

import Config

config :books,
  ecto_repos: [Books.Repo]


config :books, Books.Repo,
  database: "books",
  username: "booker",
  password: "bookerpass",
  hostname: "localhost"
```


### Ecto Repo

- Ecto ни генерира Repo (`Books.Repo`) което ни дава начин да се свържем с базата данни.
- То представлява процес, който включваме към application-а си.
- Също така задава адаптера към базата данни.

```elixir
defmodule Books.Repo do
  use Ecto.Repo,
    otp_app: :books,
    adapter: Ecto.Adapters.Postgres
end
```


### Ecto Repo

- Можем да го използваме да пишем заявки към базата.

```elixir
Books.Repo.query("SELECT 1")
#=> {:ok,
#=>  %Postgrex.Result{
#=>    command: :select,
#=>    columns: ["?column?"],
#=>    rows: [[1]],
#=>    num_rows: 1,
#=>    connection_id: 2860814,
#=>    messages: []
#=>  }}
```


### Ecto и миграции

```elixir
  def up do
    create table(:books) do
      add :isbn,        :string, size: 32
      add :title,       :string
      add :description, :string
      add :author_name, :string
      add :year,        :integer
      add :language,    :string

      timestamps()
    end
  end


  def down do
    drop table(:books)
  end
```


### Ecto и миграции

- Добра идея е да си добавим още една миграция с уникален индекс по isbn:

```elixir
  def change do
    create unique_index(:books, [:isbn], name: :books_isbn_unique_idx)
  end
```


### Ecto и заявки : insert_all

- Този тип заявки могат да запишат в базата един или много редове:

```elixir
  def insert_book!(isbn, title, description, author_name, year, language) do
    {1, [%{id: id}]} =
      Repo.insert_all(
        "books",
        [
          [
            isbn: isbn,
            title: title,
            description: description,
            author_name: author_name,
            year: year,
            language: language,
            inserted_at: DateTime.utc_now(),
            updated_at: DateTime.utc_now()
          ]
        ],
        returning: [:id]
      )

    id
  end
```


### Ecto и заявки : all

- С `Repo.all` можем да правим `SELECT` заявки:

```elixir
  def books_by_year(year) do
    query =
      from("books",
        where: [year: ^year],
        select: [:title, :author_name]
      )

    Repo.all(query)
  end
```


### Ecto и заявки : all

- `Repo.all` също така поддържа всичко, нужно на `SELECТ`, като `group_by`, `limit`, `join` и други:

```elixir
  def books_by_years_or_authors(years, authors) do
    query =
      from(b in "books",
        where: b.year in ^years or b.author_name in ^authors,
        select: [:id, :isbn, :title, :author_name, :year, :description]
      )

    Repo.all(query)
  end
```


### Ecto и заявки : one

- Функцията `Repo.one` се ползва ако задължително очакваме един резултат максимално:

```elixir
  def book_by_id(id) do
    query =
      from("books",
        where: [id: ^id],
        select: [:id, :title, :author_name, :year]
      )

    Repo.one(query)
  end
```


### Ecto и заявки : update_all

- Можем да използваме цялата сила на езика за заявки за да направим `UPDATE`:

```elixir
  def update_book_description!(book_id, description) do
    query = from(b in "books", where: b.id == ^book_id)

    {1, _} = Repo.update_all(query, set: [description: description])
  end
```


### Ecto и заявки : delete_all

- Това се отнася и за `DELETE`:

```elixir
  def delete_book_by_isbn!(isbn) do
    query = from(b in "books", where: b.isbn == ^isbn)

    {1, _} = Repo.delete_all(query)
  end
```


## Ecto схеми : с миграции

- Ако си дефинираме миграция за `countries` така:

```elixir
  def change do
    create table(:countries) do
      add :name, :string, size: 32, null: false
      add :code, :string, size: 2, null: false

      timestamps()
    end

    create unique_index(:countries, [:name])
    create unique_index(:countries, [:code])
  end
```


## Ecto схеми : с миграции

- Съответната схема би била:

```elixir
defmodule Books.Country do
  use Ecto.Schema

  import Ecto.Changeset

  schema "countries" do
    field(:name, :string)
    field(:code, :string)

    timestamps()
  end
end
```



### Ecto changeset

- Пример за `changeset` за схемата `Country`:

```elixir
  def changeset(%__MODULE__{} = country, attrs) do
    country
    |> cast(attrs, [:name, :code])
    |> validate_required([:name, :code])
    |> validate_length(:code, min: 2, max: 2)
    |> validate_length(:name, min: 4, max: 64)
    |> unique_constraint(:code)
    |> unique_constraint(:name)
    |> validate_code()
    |> validate_name()
  end
```


### Ecto changeset

- Пример за специфична валидация:

```elixir
  defp validate_code(%Ecto.Changeset{valid?: false} = changeset), do: changeset

  defp validate_code(changeset) do
    code = get_field(changeset, :code)

    if code != String.upcase(code) do
      add_error(changeset, :code, "Country code has to be in uppercase!")
    else
      changeset
    end
  end
```


### Ecto changeset и бази данни

- Невалиден `changeset` не може да се запази в базата, даже няма да има заявка:

```elixir
changeset = Country.changeset(%Country{}, %{name: "Bulgaria", code: "bg"})

{:error, changeset} = Repo.insert(changeset)

changeset.valid?
#=> false

changeset.errors
#=> [code: {"Country code has to be in uppercase!", []}]
```


## Ecto асоциации

- Нека имаме миграцията:

```elixir
  def change do
    create table(:authors) do
      add :first_name, :string, size: 32, null: false
      add :last_name,  :string, size: 32, null: false
      add :birth_date, :date
      add :country_id, references(:countries), null: false

      timestamps()
    end

    create index(:authors, [:first_name, :last_name])
  end
```


### Ecto асоциации

- Схемата `Author` представлява:

```elixir
  schema "authors" do
    field(:first_name, :string)
    field(:last_name, :string)
    field(:birth_date, :date)

    field(:name, :string, virtual: true)

    belongs_to(:country, Books.Country)

    timestamps()
  end
```


### Ecto асоциации

- Схемата `Country` се променя на:

```elixir
  schema "countries" do
    field(:name, :string)
    field(:code, :string)

    has_many :authors, Books.Author

    timestamps()
  end
```


### Ecto асоциации : *many_to_many*

- За many_to_many релацията има по добър начин за изразяване:

```elixir
  def change do
    create table(:authors_books) do
      add :author_id, references(:authors), null: false
      add :book_id, references(:books), null: false
    end
  end

# Books.Book:
  many_to_many :authors, Author, join_through: "authors_books"

# Books.Author:
  many_to_many :books, Book, join_through: "authors_books"
```


### Ecto асоциации и changesets

- Можем просто да задаваме id-тата на асоцияциите:

```elixir
  def changeset(%__MODULE__{} = author, %{country_id: _} = attrs) do
    author
    |> cast(attrs, [:first_name, :last_name, :birth_date, :country_id])
    |> validate_required([:first_name, :last_name, :country_id])
    |> foreign_key_constraint(:country_id)
    |> validate()
  end

  defp validate(changeset) do
    changeset
    |> validate_length(:first_name, min: 1, max: 32)
    |> validate_length(:last_name, min: 1, max: 32)
  end
```


### Ecto асоциации и changesets

- Можем и да задаваме целите асоциации:

```elixir
  def changeset(%__MODULE__{} = author, %{country: country} = attrs) do
    author
    |> cast(attrs, [:first_name, :last_name, :birth_date])
    |> validate_required([:first_name, :last_name])
    |> check_country(country)
    |> validate()
  end
```


### Ecto асоциации и changesets

- Важното е да проверим дали асоциацията вече съществува ако има *unique constraint*:

```elixir
  defp check_country(%{valid?: false} = changeset, _), do: changeset

  defp check_country(changeset, country) do
    country_changeset = Country.changeset(%Country{}, country)

    if country_changeset.valid? do
      put_country(Repo.get_by(Country, name: get_field(country_changeset, :name)), changeset)
    else
      changeset
    end
  end
```


### Ecto асоциации и changesets

- След, което да я запишем или само маркираме:

```elixir
  defp put_country(nil, changeset) do
    cast_assoc(changeset, :country, required: true, with: &Books.Country.changeset/2)
  end

  defp put_country(country, changeset) do
    put_assoc(changeset, :country, country)
  end
```


### Ecto асоциации и заявки

- По подразбиране асоциацията не се зарежда при `SELECT`:

```elixir
%Books.Author{
  id: 68,
  first_name: "Иван",
  last_name: "Александров",
  birth_date: ~D[1996-05-13],
  name: nil,
  country_id: 66,
  country: #Ecto.Association.NotLoaded<association :country is not loaded>,
  books: #Ecto.Association.NotLoaded<association :books is not loaded>,
  ...
}
```


### Ecto асоциации и заявки

- Можем да заредим асоциациите си с `preload`:

```elixir
  author = Books.Repo.preload(author, :country)

  %Country{name: "Bulgaria", code: "BG", id: ^country_id} = author.country
```


### Ecto асоциации и заявки

- Можем да задаваме preload и в заявки, елиминирайки *N+1* проблема:

```elixir
  def by_country_name(country_name) do
    query = from(
      a in __MODULE__,
      join: c in Country,
      on: a.country_id == c.id,
      where: c.name == ^country_name,
      preload: [:country]
    )

    query = from(a in query, select_merge: %{name: fragment("? || ' ' || ?", a.first_name, a.last_name)})

    Repo.all(query)
  end
```


## Ecto changeset без schema

```elixir
data = %{}
types = %{name: :string, email: :string}

changeset =
  {data, types}
  |> Ecto.Changeset.cast(params["sign_up"], Map.keys(types))
  |> validate_required(...)
  |> validate_length(...)

```


## Ecto changeset с embedded_schema

- Със схема, но без таблица свързана с нея:

```elixir
defmodule SignUp do
  @mail_regex ~r/^[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Za-z]{2,4}$/

  embedded_schema do
    field(:name, :string)
    field(:email, :string)
  end

  def changeset(sign_up, attrs \\ %{}) do
    sign_up
    |> cast(attrs, [:name, :email])
    |> validate_required([:name, :email])
    |> validate_length(:name, min: 4, max: 64)
    |> validate_format(field, @mail_regex)
  end
end
```


## Ecto транзакции

- Примерна транзакция за запис на книга:

```elixir
  {:ok, book_id} =
    Books.Repo.transaction(fn ->
      {:ok, %{id: country_id}} = Books.insert_or_get_country("USA", "US")

      {:ok, %{id: author_id}} =
        Books.insert_or_get_author(
          "Дан",
          "Симънс",
          Date.from_iso8601!("1948-04-04"),
          country_id
        )

      book_id = Books.insert_book!("978-619-152-344-3", "Ужас", @desc, 2013, "български")
      :ok = Books.add_authors_to_book(book_id, [author_id])

      book_id
    end)
```


### Ecto транзакции

```elixir
  {:ok, %{book_id: book_id}} =
    Multi.new()
    |> Multi.run(:country, fn _repo, _current_state ->
      {:ok, %Books.Country{}} = Books.insert_or_get_country("USA", "US")
    end)
    |> Multi.run(:author, fn _repo, %{country: %{id: country_id}} ->
      {:ok, %{id: author_id}} =
        Books.insert_or_get_author(
          "Дан",
          "Симънс",
          Date.from_iso8601!("1948-04-04"),
          country_id
        )
    end)
    |> Multi.run(:book_id, fn _repo, _current_state ->
      {:ok, Books.insert_book!("978-619-152-344-3", "Ужас", @desc, 2013, "български")}
    end)
    |> Multi.run(:authors_books, fn _repo, %{book_id: book_id, author: %{id: author_id}} ->
      case Books.add_authors_to_book(book_id, [author_id]) do
        :ok ->
          {:ok, nil}

        error ->
          error
      end
    end)
    |> Books.Repo.transaction()
```


### Ecto транзакции

- Пример за `Ecto.Multi` и операции в транзакция, базирани на списък:

```elixir
  def add_authors_to_book(book_id, author_ids) do
    steps =
      Enum.reduce(author_ids, Ecto.Multi.new(), fn author_id, steps ->
        name = "step_#{book_id}_#{author_id}"
        Ecto.Multi.insert(steps, name, %AuthorBook{book_id: book_id, author_id: author_id})
      end)

    case Repo.transaction(steps) do
      {:ok, _} ->
        :ok

      {:error, _, changeset, _changes_so_far} ->
        {:error, changeset}
    end
  end
```


## Тестване с Ecto

- Първо се нуждаем от конфигурация за тестване (*config/test.exs*).
- Важно е да сложим за *pool* `Ecto.Adapters.SQL.Sandbox`, така базата ще бъде чистена на всеки тест:

```elixir
import Config

config :books, Books.Repo,
  pool: Ecto.Adapters.SQL.Sandbox,
  database: "books_test",
  username: "booker",
  password: "bookerpass",
  hostname: "localhost"
```


### Тестване с Ecto

- В *test/test_helper.exs* дефинираме mode-ът на `Sandbox`:

```elixir
ExUnit.start()

Ecto.Adapters.SQL.Sandbox.mode(Books.Repo, :manual)
```


### Тестване с Ecto

- Сега можем да си напишем `Books.RepoCase`:

```elixir
# In test/support/repo_case.ex

defmodule Books.RepoCase do
  use ExUnit.CaseTemplate

  using do
    quote do
      alias Books.Repo

      import Ecto
      import Ecto.Query
      import Books.RepoCase
    end
  end

  setup tags do
    :ok = Ecto.Adapters.SQL.Sandbox.checkout(Books.Repo)

    unless tags[:async] do
      Ecto.Adapters.SQL.Sandbox.mode(Books.Repo, {:shared, self()})
    end

    :ok
  end
end
```


### Тестване с Ecto

- За да се компилира `RepoCase` ще трябва да променим *mix.exs*:

```elixir
  def project do
    [
      app: :books,
      version: "0.1.0",
      elixir: "~> 1.14",
      start_permanent: Mix.env() == :prod,
      elixirc_paths: elixirc_paths(Mix.env()),
      aliases: aliases(),
      deps: deps()
    ]
  end
```


### Тестване с Ecto

- За да се компилира `RepoCase` ще трябва да променим *mix.exs*:

```elixir
  defp aliases do
    [
      test: ["ecto.create --quiet", "ecto.migrate", "test"]
    ]
  end

  defp elixirc_paths(:test), do: ["lib", "test/support"]
  defp elixirc_paths(_), do: ["lib"]
```


### Тестване с Ecto

- Сега можем да си дефинираме DB тестове:

```elixir
defmodule BooksTest do
  use Books.RepoCase

  test "schemaless queries" do
    insert_test_books!()

    books = Books.books_by_year(2013)
    assert books == []

    #...
  end
end
```